import csv

from .models import Row

def get_processed_rows(filepath:str) -> list[Row]:
    i = 0
    rows = []

    with open(filepath, "r") as f:
        reader = csv.reader(f)

        for row in reader:
            i += 1
            if i == 1:
                continue

            rows.append(Row(*row))

    return rows


# From https://github.com/peliot/XIRR-and-XNPV/blob/master/financial.py

from scipy import optimize 

def secant_method(tol, f, x0):
    """
    Solve for x where f(x)=0, given starting x0 and tolerance.
    
    Arguments
    ----------
    tol: tolerance as percentage of final result. If two subsequent x values are with tol percent, the function will return.
    f: a function of a single variable
    x0: a starting value of x to begin the solver
    Notes
    ------
    The secant method for finding the zero value of a function uses the following formula to find subsequent values of x. 
    
    x(n+1) = x(n) - f(x(n))*(x(n)-x(n-1))/(f(x(n))-f(x(n-1)))
    
    Warning 
    --------
    This implementation is simple and does not handle cases where there is no solution. Users requiring a more robust version should use scipy package optimize.newton.
    """

    x1 = x0*1.1
    while (abs(x1-x0)/abs(x1) > tol):
        x0, x1 = x1, x1-f(x1)*(x1-x0)/(f(x1)-f(x0))
    return x1

def xnpv(rate,cashflows):
    """
    Calculate the net present value of a series of cashflows at irregular intervals.
    Arguments
    ---------
    * rate: the discount rate to be applied to the cash flows
    * cashflows: a list object in which each element is a tuple of the form (date, amount), where date is a python datetime.date object and amount is an integer or floating point number. Cash outflows (investments) are represented with negative amounts, and cash inflows (returns) are positive amounts.
    
    Returns
    -------
    * returns a single value which is the NPV of the given cash flows.
    Notes
    ---------------
    * The Net Present Value is the sum of each of cash flows discounted back to the date of the first cash flow. The discounted value of a given cash flow is A/(1+r)**(t-t0), where A is the amount, r is the discout rate, and (t-t0) is the time in years from the date of the first cash flow in the series (t0) to the date of the cash flow being added to the sum (t).  
    * This function is equivalent to the Microsoft Excel function of the same name. 
    """

    chron_order = sorted(cashflows, key = lambda x: x[0])
    t0 = chron_order[0][0] #t0 is the date of the first cash flow

    return sum([cf/(1+rate)**((t-t0).days/365.0) for (t,cf) in chron_order])

def xirr(cashflows,guess=0.1):
    """
    Calculate the Internal Rate of Return of a series of cashflows at irregular intervals.
    Arguments
    ---------
    * cashflows: a list object in which each element is a tuple of the form (date, amount), where date is a python datetime.date object and amount is an integer or floating point number. Cash outflows (investments) are represented with negative amounts, and cash inflows (returns) are positive amounts.
    * guess (optional, default = 0.1): a guess at the solution to be used as a starting point for the numerical solution. 
    Returns
    --------
    * Returns the IRR as a single value
    
    Notes
    ----------------
    * The Internal Rate of Return (IRR) is the discount rate at which the Net Present Value (NPV) of a series of cash flows is equal to zero. The NPV of the series of cash flows is determined using the xnpv function in this module. The discount rate at which NPV equals zero is found using the secant method of numerical solution. 
    * This function is equivalent to the Microsoft Excel function of the same name.
    * For users that do not have the scipy module installed, there is an alternate version (commented out) that uses the secant_method function defined in the module rather than the scipy.optimize module's numerical solver. Both use the same method of calculation so there should be no difference in performance, but the secant_method function does not fail gracefully in cases where there is no solution, so the scipy.optimize.newton version is preferred.
    """
    
    #return secant_method(0.0001,lambda r: xnpv(r,cashflows),guess)
    return optimize.newton(lambda r: xnpv(r,cashflows),guess)


import datetime
from decimal import Decimal

from dateutil import parser

from .utils import xirr
from .models import Row
from .csv_parser import get_processed_rows


def get_ticker_value(ticker:str) -> Decimal:
    values = {
        "VUSA": Decimal("6533.46"),
    }
    return values[ticker]


def get_total_shares(ticker:str, rows:list[Row]) -> Decimal:
    total_shares = Decimal("0")

    for row in rows:
        if row.ticker == ticker:
            if "buy" in row.action:
                total_shares += Decimal(row.num_shares)
            if "sell" in row.action:
                total_shares -= Decimal(row.num_shares)

    return total_shares


def get_all_tickers_with_total_shares(rows:list[Row]) -> dict:
    tickers = {}

    for row in rows:
        if row.ticker not in tickers:
            tickers[row.ticker] = get_total_shares(row.ticker, rows)

    return tickers


def get_xirr_for_ticker(ticker:str, rows:list[Row]) -> float:
    cashflows = []

    for row in rows:
        if row.ticker == ticker:
            if row.action == "Market buy":
                cashflows.append((parser.parse(row.time).date(), -1 * float(Decimal(row.total))))
            if row.action == "Market sell":
                cashflows.append((parser.parse(row.time).date(), float(Decimal(row.total))))

    cashflows.append((datetime.datetime.now().date(), float(get_ticker_value(ticker))))
    ticker_xirr = xirr(cashflows)
    return ticker_xirr


def get_total_cashflows(rows: list[Row], current_value: Decimal) -> list[tuple]:
    cashflows = [(parser.parse(row.time).date(), -1 * float(Decimal(row.total))) for row in rows if row.action == "Deposit"]
    cashflows.append((datetime.datetime.now().date(), float(current_value)))
    return cashflows


def get_summary(filepath: str, current_value: Decimal) -> list[list]:
    rows = get_processed_rows(filepath)
    total_deposits = sum(
        [Decimal(row.total) for row in rows if row.action == "Deposit"]
    )
    cashflows = get_total_cashflows(rows, current_value)
    returns = xirr(cashflows)
    table = [
        ["Total Deposits", f"£{'{:,}'.format(total_deposits)}"],
        ["Current Portfolio Value", f"£{'{:,}'.format(current_value)}"],
        ["Portfolio Returns", f"£{'{:,}'.format(current_value - total_deposits)} ({(current_value - total_deposits)*100/total_deposits:.2f}%)"],
        ["Portfolio XIRR", f"{returns*100:.2f}%"],
    ]
    return table


# table2 = [
#     ["VUSA", f"{get_xirr_for_ticker('VUSA', rows)*100:.2f}%", f"{get_total_shares('VUSA', rows)}"],
# ]
# print(tabulate(table2, ["Ticker", "XIRR", "Total Shares Owned"], tablefmt="fancy_grid"))
# print(get_all_tickers_with_total_shares(rows))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['portfolio_analyzer']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'scipy>=1.7.1,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'yfinance>=0.1.63,<0.2.0']

entry_points = \
{'console_scripts': ['portfolio = portfolio_analyzer.cli:main']}

setup_kwargs = {
    'name': 'portfolio-analyzer',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Tapan Pandita',
    'author_email': 'tapan.pandita@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from dataclasses import dataclass


@dataclass
class Row:
    action: str
    time: str
    isin: str
    ticker: str
    name: str
    num_shares: str
    price_per_share: str
    currency_pps: str
    exchange_rate: str
    result: str
    total: str
    withholding_tax: str
    currency_withholding_tax: str
    charge_amount: str
    transaction_fee: str
    finra_fee: str
    notes: str
    id: str
    currency_conversion_fee: str


from decimal import Decimal

import click
from tabulate import tabulate

from .analyzer import get_summary


@click.group()
def cli():
    pass


@cli.command()
@click.option("--path", "-p", help="path to the CSV file with data", required=True)
@click.option("--value", "-v", help="current portfolio value", required=True, type=Decimal)
def summary(path, value):
    summary = get_summary(path, value)
    click.echo(tabulate(summary, tablefmt="fancy_grid"))


@cli.command()
@click.option("--path", "-p", help="path to the CSV file with data")
@click.option("--tickers", "-t", default="", help="filter to only these tickers, leave empty for all")
def details(path, tickers):
    click.echo(path)
    click.echo(tickers)


def main():
    cli()